#import "BTDevice.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#import <CoreFoundation/CoreFoundation.h>
#import <CoreGraphics/CGContext.h>
#import <UIKit/UIKit.h>
#import <UIKit/UITextView.h>
#import <UIKit/UIFont.h>
#import <NSGraphicsContext.h>
#import "CocoaTouch.h"

#include <btstack/btstack.h>
#include <btstack/run_loop.h>
#include <btstack/hci_cmds.h>
#include <BTstackManager.h>

int main(int argc, char *argv[]) {
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
    int retVal = UIApplicationMain(argc, argv, @"CocoaTouch", @"CocoaTouch");
	[pool release];
	return retVal;
}

@implementation CocoaTouch

-(void) rfcommConnectionCreatedAtAddress:(bd_addr_t)addr forChannel:(uint16_t)channel asID:(uint16_t)connectionID{
	//rfStat.text = @"connected";
	if (connectionID == 1) {
    	[rfStat performSelectorOnMainThread:@selector(setText:)
            withObject: @"debut" 
            waitUntilDone:YES];
	} else {
    	[rfStat performSelectorOnMainThread:@selector(setText:)
            withObject: @"coucou" 
            waitUntilDone:YES];
	}
}
-(void) rfcommConnectionCreateFailedAtAddress:(bd_addr_t)addr forChannel:(uint16_t)channel error:(BTstackError)error{
}
-(void) rfcommConnectionClosedForConnectionID:(uint16_t)connectionID{
}
-(void) rfcommDataReceivedForConnectionID:(uint16_t)connectionID withData:(uint8_t *)packet ofLen:(uint16_t)size{
    [rfStat performSelectorOnMainThread:@selector(setText:)
            withObject: [[NSString alloc] initWithCString:(char*)packet encoding : 1]
            waitUntilDone:YES];
	//txtHeure.text = @"recut";
}

-(void) activatedBTstackManager:(BTstackManager*) manager {
    NSLog(@"activated!");
//    [bt startDiscovery];
}

-(void) btstackManager:(BTstackManager*)manager activationFailed:(BTstackError)error {
    NSLog(@"activationFailed error 0x%02x!", error);
};
-(void) discoveryInquiryBTstackManager:(BTstackManager*) manager {
    NSLog(@"discoveryInquiry!");
    [bt storeDeviceInfo];
}
-(void) discoveryStoppedBTstackManager:(BTstackManager*) manager {
    NSLog(@"discoveryStopped!");
}
-(void) btstackManager:(BTstackManager*)manager discoveryQueryRemoteName:(int)deviceIndex {
    NSLog(@"discoveryQueryRemoteName %u/%u!", deviceIndex+1, [bt numberOfDevicesFound]);
}
-(void) btstackManager:(BTstackManager*)manager deviceInfo:(BTDevice*)device{
    NSLog(@"Device Info: addr %@ name %@ COD 0x%06x", [device addressString], [device name], [device classOfDevice] );
}


/* Event when a device is selected */
-(BOOL) discoveryView:(BTDiscoveryViewController*)discoveryView willSelectDeviceAtIndex:(int)deviceIndex {
    if (selectedDevice) return NO;
    selectedDevice = [bt deviceAtIndex:deviceIndex];
    BTDevice *device = selectedDevice;
    [bt stopDiscovery];
	BTstackError err = [bt createRFCOMMConnectionAtAddress:[device address] withChannel:11 authenticated:NO];
	[nav.view removeFromSuperview];
	//[bt removeListener:inqView];
	if (err != 0)
		txtHeure.text = @"error";
	else{
		ready = 1;
		txtHeure.text = [device toString];//@"coucou";
	}
    return NO;
}
-(void) statusCellSelectedDiscoveryView:(BTDiscoveryViewController*)discoveryView {
    if (![bt isDiscoveryActive]) {
        selectedDevice = nil;
        [bt startDiscovery];
    }
    NSLog(@"statusCellSelected!");
}

- (void)applicationDidFinishLaunching:(UIApplication *)application {    
	ready = 0;
	selectedDevice = nil;
	// create BTInquiryView
	inqView = [[BTDiscoveryViewController alloc] init];
	[inqView setDelegate:self];
    // Override point for customization after application launch
	nav = [[UINavigationController alloc] initWithRootViewController:inqView];

	window = [[UIWindow alloc] initWithFrame:[[UIScreen mainScreen] bounds]];
	window.backgroundColor = [UIColor whiteColor];
    CGRect txtFrame1 = CGRectMake(0,50,250,150);
    CGRect txtFrame2 = CGRectMake(0,100,250,150);
    UIFont *font = [UIFont boldSystemFontOfSize:18.0];

	txtHeure = [[UITextField alloc] initWithFrame:txtFrame1];
	txtHeure.text = @"testGPS";
    txtHeure.font = font;

	rfStat = [[UITextField alloc] initWithFrame:txtFrame2];
	rfStat.text = @"testGPS";
    rfStat.font = font;

	[window addSubview:txtHeure];
	[window addSubview:rfStat];
	[window addSubview:nav.view];
	[window makeKeyAndVisible];


	// btstack
	bt = [BTstackManager sharedInstance];
	[bt setDelegate:self];
	[bt addListener:self];
	[bt addListener:inqView];
		

	BTstackError err = [bt activate];
}

- (void)dealloc {
    //[window release];
    [super dealloc];
}

-(void) btstackManager:(BTstackManager*) manager handlePacketWithType:(uint8_t) packet_type forChannel:(uint16_t) channel
               andData:(uint8_t *)packet withLen:(uint16_t) size {
}
@end
