#import "BTDevice.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#import <CoreFoundation/CoreFoundation.h>
#import <CoreGraphics/CGContext.h>
#import <UIKit/UIKit.h>
#import <UIKit/UITextView.h>
#import <UIKit/UIFont.h>
#import <NSGraphicsContext.h>
#import "michelin.h"

#include <btstack/btstack.h>
#include <btstack/run_loop.h>
#include <btstack/hci_cmds.h>
#include <BTstackManager.h>

int main(int argc, char *argv[]) {
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
    int retVal = UIApplicationMain(argc, argv, @"michelin", @"michelin");
	[pool release];
	return retVal;
}

@implementation michelin
char tmp[2048];
NSFileHandle *fileHandle;
int nb=0;
-(void) rfcommDataReceivedForConnectionID:(uint16_t)connectionID withData:(uint8_t *)packet ofLen:(uint16_t)size{
	memcpy(tmp+nb,packet,size);
	nb+=size;
	tmp[nb]='\0';
	if ((packet[size-1] == '\n')){
		NSString *date = [formatDate stringFromDate:[NSDate date]];
		NSString *str = [[NSString alloc] initWithCString:(char*)tmp encoding : NSASCIIStringEncoding];
		/* preparation et ecriture dans le fichier */
		NSString *file_content = [[NSString alloc] initWithFormat:@"%@ %@", date, str];
		NSData *data = [date dataUsingEncoding:NSASCIIStringEncoding];
		[fileHandle writeData:data];
		//[fileHandle synchronizeFile];
		/* preparation pour l'affichage de la courbe */
		NSArray *content = [str componentsSeparatedByCharactersInSet:
								[NSCharacterSet whitespaceCharacterSet]];
		NSString *v1 = [content objectAtIndex:1];
		NSString *v5 = [content objectAtIndex:5];
		CGFloat diff_val = ([v5 doubleValue] - [v1 doubleValue])/10000;
		/* affichage dans les texts view */
    	[date_heure performSelectorOnMainThread:@selector(setText:)
       							withObject: date waitUntilDone:YES];
    	[valDiff performSelectorOnMainThread:@selector(setText:)
    	        withObject: [[NSString alloc] initWithFormat:@"%f",diff_val]
       		     waitUntilDone:YES];
		/* mise a jour de la courbe */
		[affSin addPoint: diff_val];
		/* nettoyage et remise a zero */
		[file_content release];
		[str release];
		[date release];
		nb = 0;
	}
}



/* Event when a device is selected */
-(BOOL) discoveryView:(BTDiscoveryViewController*)discoveryView willSelectDeviceAtIndex:(int)deviceIndex {
    if (selectedDevice) return NO;
    selectedDevice = [bt deviceAtIndex:deviceIndex];
    BTDevice *device = selectedDevice;
    [bt stopDiscovery];
	[bt createRFCOMMConnectionAtAddress:[device address] withChannel:11 authenticated:NO];
	[nav.view removeFromSuperview];
	//[bt removeListener:inqView];
    return NO;
}
-(void) statusCellSelectedDiscoveryView:(BTDiscoveryViewController*)discoveryView {
    if (![bt isDiscoveryActive]) {
        selectedDevice = nil;
        [bt startDiscovery];
    }
    NSLog(@"statusCellSelected!");
}

- (void)applicationDidFinishLaunching:(UIApplication *)application {    
	selectedDevice = nil;
	/* formatage de la date */
	formatDate = [[NSDateFormatter alloc] init];
	[formatDate setDateFormat:@"yyyy_MM_dd__HH_MM_ss"];
	NSString *currentDate = [formatDate stringFromDate:[NSDate date]];
	/* creation du fichier de sauvegarde */
	file_path = [[NSString alloc] initWithFormat:@"/var/mobile/Documents/michelin_%@.txt",
					currentDate];
	[currentDate release];
	[[NSFileManager defaultManager] createFileAtPath:file_path
									contents:@"" attributes:nil]; 
    fileHandle  = [[NSFileHandle fileHandleForWritingAtPath:file_path]retain];
	/* creation interface graphique */
	window = [[UIWindow alloc] initWithFrame:[[UIScreen mainScreen] bounds]];
	// create BTInquiryView
	inqView = [[BTDiscoveryViewController alloc] init];
	[inqView setDelegate:self];
    // Override point for customization after application launch
	nav = [[UINavigationController alloc] initWithRootViewController:inqView];


    CGRect windowRect = [ [ UIScreen mainScreen ] applicationFrame ];
    windowRect.origin.x = windowRect.origin.y = 0.0f;
	/* affSin part */
    CGRect imgRect2 = CGRectMake(0,0,windowRect.size.width,
								windowRect.size.height);
    affSin = [[afficheSinu alloc] initWithFrame: imgRect2];
	[affSin setLineWidth:5];
	[affSin setBackground: [UIColor whiteColor]];
	[affSin setForeground: [UIColor blackColor]];
	affSin.nbPoints=100;

	/* UITextView */
    UIFont *font = [UIFont boldSystemFontOfSize:18.0];
	CGAffineTransform rotation = CGAffineTransformRotate(CGAffineTransformIdentity, 
						(90.0/180.0 * M_PI));
	/* date_heure */
    CGRect date_heure_frame = CGRectMake(110,65,250,150);
    date_heure = [[UILabel alloc] initWithFrame:date_heure_frame];
	//date_heure.textColor = black;
    date_heure.text = @"michelin";
    date_heure.font = font;
	date_heure.transform = rotation;
	/* diff_affiche */
    CGRect diff_frame = CGRectMake(90,65,250,150);
    valDiff = [[UILabel alloc] initWithFrame:diff_frame];
	//valDiff.textColor = black;
    valDiff.text = @"valDiff";
    valDiff.font = font;
	valDiff.transform = rotation;


    [window addSubview:affSin];
    [window addSubview:date_heure];
    [window addSubview:valDiff];
	[window addSubview:nav.view];
	[window makeKeyAndVisible];

	// btstack
	bt = [BTstackManager sharedInstance];
	[bt setDelegate:self];
	[bt addListener:self];
	[bt addListener:inqView];
	[bt activate];
	
}

- (void)dealloc {
    [window release];
    [super dealloc];
}
-(void) activatedBTstackManager:(BTstackManager*) manager {}
-(void) btstackManager:(BTstackManager*)manager activationFailed:(BTstackError)error {};
-(void) discoveryInquiryBTstackManager:(BTstackManager*) manager {
    [bt storeDeviceInfo];
}
-(void) discoveryStoppedBTstackManager:(BTstackManager*) manager {}
-(void) btstackManager:(BTstackManager*)manager discoveryQueryRemoteName:(int)deviceIndex {}
-(void) btstackManager:(BTstackManager*)manager deviceInfo:(BTDevice*)device{}

-(void) btstackManager:(BTstackManager*) manager handlePacketWithType:(uint8_t) packet_type forChannel:(uint16_t) channel
               andData:(uint8_t *)packet withLen:(uint16_t) size {}
-(void) rfcommConnectionCreatedAtAddress:(bd_addr_t)addr forChannel:(uint16_t)channel asID:(uint16_t)connectionID{}
-(void) rfcommConnectionCreateFailedAtAddress:(bd_addr_t)addr forChannel:(uint16_t)channel error:(BTstackError)error{}
-(void) rfcommConnectionClosedForConnectionID:(uint16_t)connectionID{}
@end
